﻿using HeuristicLab.Algorithms.Bandits.BanditPolicies;
using HeuristicLab.Algorithms.GeneticProgramming;
using HeuristicLab.Algorithms.GrammaticalOptimization;
using HeuristicLab.Algorithms.MonteCarloTreeSearch;
using HeuristicLab.Problems.GrammaticalOptimization;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Problems.GrammaticalOptimization.SymbReg;

namespace Evaluation.ViewModel
{
    public class EvaluationViewModel : INotifyPropertyChanged
    {
        private List<ISymbolicExpressionTreeProblem> problems;

        public List<ISymbolicExpressionTreeProblem> Problems
        {
            get { return this.problems; }
            set
            {
                this.problems = value;
                this.OnPropertyChanged("Problems");
            }
        }

        private ISymbolicExpressionTreeProblem selectedProblem;

        public ISymbolicExpressionTreeProblem SelectedProblem
        {
            get { return this.selectedProblem; }
            set
            {
                this.selectedProblem = value;
                this.OnPropertyChanged("SelectedProblem");
            }
        }

        private List<Type> algorithms;

        public List<Type> Algorithms
        {
            get { return this.algorithms; }
            set
            {
                this.algorithms = value;
                this.OnPropertyChanged("Algorithms");
            }
        }

        private Type selectedAlgorithm;

        public Type SelectedAlgorithm
        {
            get { return this.selectedAlgorithm; }
            set
            {
                this.selectedAlgorithm = value;
                this.OnPropertyChanged("SelectedAlgorithm");
            }
        }

        private List<Type> policies;

        public List<Type> Policies
        {
            get { return this.policies; }
            set
            {
                this.policies = value;
                this.OnPropertyChanged("Policies");
            }
        }

        private Type selectedPolicy;

        public Type SelectedPolicy
        {
            get { return this.selectedPolicy; }
            set { this.selectedPolicy = value; this.OnPropertyChanged("SelectedPolicy"); }
        }

        private double bestKnownQuality;

        public double BestKnownQuality
        {
            get { return this.bestKnownQuality; }
            set
            {
                this.bestKnownQuality = value;
                this.OnPropertyChanged("BestKnownQuality");
            }
        }

        private double currentBestQuality;

        public double CurrentBestQuality
        {
            get { return this.currentBestQuality; }
            set
            {
                this.currentBestQuality = value;
                this.OnPropertyChanged("CurrentBestQuality");
            }
        }

        private long evaluations;

        public long Evaluations
        {
            get { return this.evaluations; }
            set
            {
                this.evaluations = value;
                this.OnPropertyChanged("Evaluations");
            }
        }

        private int maxLen;

        public int MaxLen
        {
            get { return this.maxLen; }
            set { this.maxLen = value; this.OnPropertyChanged("MaxLen"); }
        }

        private int maxEvaluations;

        public int MaxEvaluations
        {
            get { return this.maxEvaluations; }
            set
            {
                this.maxEvaluations = value;
                this.OnPropertyChanged("MaxEvaluations");
            }
        }

        private string headerString;

        public string HeaderString
        {
            get { return this.headerString; }
            set
            {
                headerString = value;
                this.OnPropertyChanged("HeaderString");
            }
        }

        private string verticalAxisString;

        public string VerticalAxisString
        {
            get { return this.verticalAxisString; }
            set
            {
                verticalAxisString = value;
                this.OnPropertyChanged("VerticalAxisString");
            }
        }

        private string horizontalAxisString;

        public string HorizontalAxisString
        {
            get { return this.horizontalAxisString; }
            set
            {
                horizontalAxisString = value;
                this.OnPropertyChanged("HorizontalAxisString");
            }
        }

        private double evaluationsPerSec;

        public double EvaluationsPerSec
        {
            get { return this.evaluationsPerSec; }
            set
            {
                evaluationsPerSec = value;
                this.OnPropertyChanged("EvaluationsPerSec");
            }
        }

        private long bestSolutionFoundAt;

        public long BestSolutionFoundAt
        {
            get { return this.bestSolutionFoundAt; }
            set
            {
                bestSolutionFoundAt = value;
                this.OnPropertyChanged("BestSolutionFoundAt");
            }
        }


        public EvaluationViewModel()
        {
            InitializeProblems();
            InitializeAlgorithms();
            InitializePolicies();
        }

        private void InitializePolicies()
        {
            this.policies = new List<Type>();
            this.policies.Add(typeof(ActiveLearningPolicy));
            this.policies.Add(typeof(BernoulliThompsonSamplingPolicy));
            this.policies.Add(typeof(BoltzmannExplorationPolicy));
            this.policies.Add(typeof(EpsGreedyPolicy));
            this.policies.Add(typeof(GenericThompsonSamplingPolicy));
            this.policies.Add(typeof(ModifiedUCTPolicy));
            this.policies.Add(typeof(RandomPolicy));
            this.policies.Add(typeof(ThresholdAscentPolicy));
            this.policies.Add(typeof(UCB1Policy));
            this.policies.Add(typeof(UCB1TunedPolicy));
            this.policies.Add(typeof(UCBNormalPolicy));
            this.policies.Add(typeof(UCTPolicy));

            this.selectedPolicy = typeof(UCB1Policy);
        }

        private void InitializeAlgorithms()
        {
            this.algorithms = new List<Type>();
            this.algorithms.Add(typeof(RandomSearch));
            this.algorithms.Add(typeof(SequentialSearch));
            this.algorithms.Add(typeof(MonteCarloTreeSearch));
            this.algorithms.Add(typeof(StandardGP));
            this.algorithms.Add(typeof(OffspringSelectionGP));

            this.selectedAlgorithm = typeof(MonteCarloTreeSearch);
        }

        private void InitializeProblems()
        {
            this.problems = new List<ISymbolicExpressionTreeProblem>();
            this.problems.Add(new SymbolicRegressionPoly10Problem());
            this.problems.Add(new SantaFeAntProblem());
            this.problems.Add(new EvenParityProblem());
            this.problems.Add(new RoyalPairProblem());
            this.problems.Add(new PalindromeProblem());
            this.problems.Add(new PermutationProblem());
            this.problems.Add(new RoyalSymbolProblem());
            // TODO: still some missing...

            this.selectedProblem = problems.FirstOrDefault();
        }

        #region INotifyPropertyChanged members

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion INotifyPropertyChanged members
    }
}
